/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.ior;

import mqc.holders.*;

/**
* A tagged profile base class.
*
* @author Olav Haugan
* @author A. David McKinnon
* @author Eivind Naess
*/
public abstract class TaggedProfile
{
   public static final int IIOP = 0;
   public static final int IIOPLITE = 1;
   public static final int MQCIOP = 2;
   public static final int UIOP = 3;
   public static final int UIOPLITE = 4;
   public static final int UMQCIOP = 5;

   public int profileTag;
   public byte[] objectKey;
   public int seqLen;
   public int major;
   public int minor;
   public String host;
   public int port;

   protected boolean bigEndian;
   /**
    * Constructor. Initializes variables
    *
    * @param IORParser The ior object
    */
   public TaggedProfile(int profileTag)
   {
      this.profileTag = profileTag;
      this.bigEndian = true;
      objectKey = null;
      seqLen = 0;
   }

   /**
    * Parses an IOR string
    *
    * @param byte[] The ior buffer
    * @param int The pointer to the buffer
    * @param int The sequence length
    * @return int The new pointer to the buffer
    */
   public int parse(byte[] buf, int ptr, int seqLen)
   {
      int endOfSeq = ptr + seqLen;

      /* Check for big endian / little endian */
      if(buf[ptr++] == 1)
      {
         bigEndian = false;
      }
      else
      {
         bigEndian = true;
      }

      /* Get version number */
      major = buf[ptr++];
      minor = buf[ptr++];

      ptr = (ptr + 3) & ~3;
      int hostLen = 0;
      if(bigEndian)
      {
         hostLen = ((buf[ptr++] & 0xFF) << 24) |
                   ((buf[ptr++] & 0xFF) << 16) |
                   ((buf[ptr++] & 0xFF) << 8) |
                   ((buf[ptr++] & 0xFF));
      }
      else
      {
         hostLen = (buf[ptr++] & 0xFF) |
                   ((buf[ptr++] & 0xFF) << 8) |
                   ((buf[ptr++] & 0xFF) << 16) |
                   ((buf[ptr++] & 0xFF) << 24);
      }

      /* Get host name */
      int i = 0;
      StringBuffer tmp = new StringBuffer();
      for(i = 0; i < hostLen-1; i++)
      {
         tmp.append( (char) buf[ptr]);
         ptr++;
      }
      ptr++;
      host = tmp.toString();
      /* Get port */
      ptr = (ptr + 1) & ~1;
      if(bigEndian)
      {
         port = (int) (((buf[ptr++] & 0xFF) << 8) | (buf[ptr++] & 0xFF));
      }
      else
      {
         port = (int) ((buf[ptr++] & 0xFF) | ((buf[ptr++] & 0xFF) << 8));
      }

      /* Get the object key length */
      ptr = (ptr + 3) & ~3;
      int objKeyLen = 0;
      if(bigEndian)
      {
         objKeyLen = ((buf[ptr++] & 0xFF) << 24) |
                     ((buf[ptr++] & 0xFF) << 16) |
                     ((buf[ptr++] & 0xFF) << 8) |
                     ((buf[ptr++] & 0xFF));
      }
      else
      {
         objKeyLen = (buf[ptr++] & 0xFF) |
                     ((buf[ptr++] & 0xFF) << 8) |
                     ((buf[ptr++] & 0xFF) << 16) |
                     ((buf[ptr++] & 0xFF) << 24);
      }

      /* Get the object key */
      objectKey = new byte[objKeyLen];
      System.arraycopy(buf, ptr, objectKey, 0, objKeyLen);
      ptr += objKeyLen;

      //We'll just ignore the context if there is any
      if(ptr != endOfSeq)
      {
         ptr = endOfSeq;
      }
      /*
        ptr = (ptr + 3) & ~3;
        
        int contextLen;
        if(bigEndian)
      {
         contextLen = (((buf[ptr++] & 0xFF) << 24) |
         ((buf[ptr++] & 0xFF) << 16) |
         ((buf[ptr++] & 0xFF) << 8) |
         (buf[ptr++] & 0xFF));
     }
     else
     {
         contextLen = ((buf[ptr++] & 0xFF) |
         ((buf[ptr++] & 0xFF) << 8) |
         ((buf[ptr++] & 0xFF) << 16) |
         ((buf[ptr++] & 0xFF) << 24));
     }
     
     int seqLen = 0;
     for(i = 0; i < contextLen; i++)
     {
        ptr = (ptr + 3) & ~3;
        ptr += 4; //ignore service Id
        if(bigEndian)
        {
           seqLen = (((buf[ptr++] & 0xFF) << 24) |
           ((buf[ptr++] & 0xFF) << 16) |
           ((buf[ptr++] & 0xFF) << 8) |
           (buf[ptr++] & 0xFF));
        }
        else
        {
           seqLen = ((buf[ptr++] & 0xFF) |
           ((buf[ptr++] & 0xFF) << 8) |
           ((buf[ptr++] & 0xFF) << 16) |
           ((buf[ptr++] & 0xFF)  << 24));
        }
     ptr += seqLen;
     }
     */
      return ptr;
   }

   /**
    * Parses a corbaloc string
    *
    * @param String The corbaloc string
    * @return int The number of bytes consumed in the string
    */
   public int parseCorbaloc(String str)
   {
      int cPtr = str.indexOf(':');
      host = str.substring(0, cPtr);

      int sPtr = str.indexOf('/', cPtr);
      String portStr = str.substring(cPtr+1, sPtr);
      port = Integer.valueOf(portStr).intValue();

      int kPtr = str.indexOf('/', sPtr+1);
      String poaStr = str.substring(sPtr+1, kPtr);
      String keyStr = str.substring(kPtr+1);
      long   keyID  = Long.parseLong(keyStr);

      //Object key consist of: Length of POA name, the poa name + key
      int poaLen = poaStr.length();
      objectKey = new byte[4 + poaLen + 4];
      objectKey[0] = (byte) (poaLen >> 24);
      objectKey[1] = (byte) (poaLen >> 16);
      objectKey[2] = (byte) (poaLen >> 8);
      objectKey[3] = (byte) poaLen;
      byte[] poaName = poaStr.getBytes();
      System.arraycopy(poaName, 0, objectKey, 4, poaLen);

      objectKey[4+poaLen] = (byte) (keyID >> 24);
      objectKey[5+poaLen] = (byte) (keyID >> 16);
      objectKey[6+poaLen] = (byte) (keyID >> 8);
      objectKey[7+poaLen] = (byte) keyID;

      return str.length();
   }

   /**
    * Finds the total size and the sequence size of the 
    * tagged profile
    *
    * @param int The current size
    * @return int The new size
    */
   public int size(int size)
   {
      //profiletag + seq. length
      size += (4 + 4);

      //endianess + major + minor + padding + hostnameLength + hostname + '\0'
      seqLen = size + 1 + 1 + 1 + 1 + 4 + host.length() + 1;
      //padding
      seqLen = (seqLen + 1) & ~1;

      //port
      seqLen += 2;

      //Padding
      seqLen = (seqLen + 3) & ~3;

      //objkeylen + objkey
      seqLen += 4 + objectKey.length;

      seqLen = (seqLen + 3) & ~3;
      seqLen += 4; //extra info...
      seqLen -= size;
      size += seqLen;
      return size;
   }

   /**
    * Encodes a tagged profile
    *
    * @param byte[] The buffer
    * @param IntHolder The current pointer
    * @return byte[] The buffer
    */
   public  byte[] encodeProfile(byte[] ior, IntHolder ptr)
   {
      //ProfileTag
      ptr.value = (ptr.value + 3) & ~3;
      ior[ptr.value++] = (byte) (profileTag >> 24);
      ior[ptr.value++] = (byte) (profileTag >> 16);
      ior[ptr.value++] = (byte) (profileTag >> 8);
      ior[ptr.value++] = (byte) (profileTag);

      //Sequence length
      ior[ptr.value++] = (byte) ((seqLen) >> 24);
      ior[ptr.value++] = (byte) ((seqLen) >> 16);
      ior[ptr.value++] = (byte) ((seqLen) >> 8);
      ior[ptr.value++] = (byte) (seqLen);

      //big endian
      ior[ptr.value++] = (byte) 0;

      //major
      ior[ptr.value++] = (byte) major;
      //minor
      ior[ptr.value++] = (byte) minor;

      //Padding
      ptr.value = (ptr.value + 3) & ~3;

      //host name length
      ior[ptr.value++] = (byte) ((host.length() +1) >> 24);
      ior[ptr.value++] = (byte) ((host.length() +1) >> 16);
      ior[ptr.value++] = (byte) ((host.length() +1) >> 8);
      ior[ptr.value++] = (byte) (host.length() +1);

      //hostname
      byte[] tmpHost = host.getBytes();
      System.arraycopy(tmpHost, 0, ior, ptr.value, tmpHost.length);
      ptr.value += tmpHost.length;
      ior[ptr.value++] = (byte) 0;

      //padding
      ptr.value = (ptr.value + 1) & ~1;

      //port
      ior[ptr.value++] = (byte) (port >> 8);
      ior[ptr.value++] = (byte) port;

      //padding
      ptr.value = (ptr.value + 3) & ~3;

      //Object key length
      ior[ptr.value++] = (byte) ((objectKey.length) >> 24);
      ior[ptr.value++] = (byte) ((objectKey.length) >> 16);
      ior[ptr.value++] = (byte) ((objectKey.length) >> 8);
      ior[ptr.value++] = (byte) (objectKey.length);

      System.arraycopy(objectKey, 0, ior, ptr.value, objectKey.length);
      ptr.value += objectKey.length;

      //padding
      ptr.value = (ptr.value + 3) & ~3;
      ptr.value += 4;

      return ior;
   }

   /**
    * Encodes a tagged profile into a corbaloc reference
    *
    * @param String The corbaloc string
    * @return String The corbaloc reference
    */
   public String encodeCorbalocProfile(String corbalocStr)
   {
      //returning only host:port/<object/key> seperation has to be done with ';'
      corbalocStr = corbalocStr + ":";
      //hostname
      corbalocStr += host + ":";

      //port
      corbalocStr += port + "/";

      //Object key
      // skip the first 4 bytes which encode the length
      // skip the last 4 bytes which encode the object key itself
      // the remaining bytes in the middle are the POA's name
      for (int i = 4;  i < objectKey.length - 4;  i++)
      {
         corbalocStr += (char) objectKey[i];
      }
      // decode the key value into an int variable
      int index = objectKey.length - 4;
      int key = (int) (((objectKey[index++] & 0xff) << 24) |
                       ((objectKey[index++] & 0xff) << 16) |
                       ((objectKey[index++] & 0xff) << 8) |
                       ((objectKey[index++] & 0xff)));
      corbalocStr += "/" + key;

      return corbalocStr;
   }

   /**
    * Clones itself
    *
    * @return this THIS
    */
   public abstract Object clone();

}
