/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.protocols;

/**
 * @author Kim Christian Swenson
 */
public class CDRInputStream
{
   private boolean bigEndian;
   private byte[] buffer;
   private int bufferPtr;
   private boolean m_align;


   /**
    * Reads an int from the deMarshalBuffer based
    * on big or little endian
    *
    * @param buffer 
    * @param ptr The pointer where to start reading
    * @param bEndian Indicates to read big or little endian from the stream
    * @return The integer value read from the stream
    */
   public static int readInt(byte [] buffer, int ptr, boolean bigEndian)
   {
      int retval=0;

      if (bigEndian)
      {
         retval = (((buffer[ptr++] & 0xFF) << 24) |
                   ((buffer[ptr++] & 0xFF) << 16) |
                   ((buffer[ptr++] & 0xFF) <<  8) |
                    (buffer[ptr++] & 0xFF)       );
      }
      else
      {
         retval =  ((buffer[ptr++] & 0xFF)        |
                   ((buffer[ptr++] & 0xFF) <<  8) |
                   ((buffer[ptr++] & 0xFF) << 16) |
                   ((buffer[ptr++] & 0xFF) << 24));
      }
      return retval;
   }

  /**
   * Default constructor.
   */
   public CDRInputStream()
   {
      this(null, 0, false);
   }

  /**
   * Constructs a new CDRInputStream with given user properties.
   *
   * @param buffer  buffer holding the stream
   * @param offset  start offset
   * @param bigEndian  indicates whether to use Big Endian when
   *                   demarshalling CDR stream
   */
   public CDRInputStream(byte[] buffer, int offset, boolean bigEndian)
   {
      this.buffer = buffer;
      bufferPtr = offset;
      this.bigEndian = bigEndian;
      m_align = true;
   }

   public int bytesStreamed()
   {
      return bufferPtr;
   }

   public void setPtr( int ptr )
   {
      bufferPtr = ptr;
   }

   public boolean isAlign()
   {
      return m_align;
   }

   public void setAlign( boolean al )
   {
      m_align = al;
   }
   
  /**
   * @return  whether this CDR stream uses BIG ENDIAN when unmarshalling
   */
   public boolean isBigEndian()
   {
      return bigEndian;
   }

   public void setBigEndian( boolean big )
   {
      bigEndian = big;
   }

  /**
   * Un-marshals a byte from the stream.
   *
   * @return  next octet in the stream
   */
   public byte readOctet()
   {
      return buffer[bufferPtr++];
   }

  /**
   * Un-marshals a boolean from the stream.
   *
   * @return  next octet interpreted as a boolean
   */
   public boolean readBoolean()
   {
      return buffer[bufferPtr++] != 0x00;
   }

  /**
   * Un-marshals a char from the stream.
   *
   * @return  next octet interpreted as a char
   */
   public char readChar()
   {
      return (char) buffer[bufferPtr++];
   }

  /**
   * Un-marshals an unsigned short from the stream.
   *
   * @return  next two octets interpreted as an unsigned short
   */
   public short readUnsignedShort()
   {
      return readShort();
   }

  /**
   * Un-marshals a short from the stream.
   *
   * @return  next two octets interpreted as a signed short
   */
   public short readShort()
   {
      //Align bufferPtr so it points to the next multiple of 2.
      bufferPtr = align(bufferPtr, 2);

      if(bigEndian)
      {
         return (short) ((buffer[bufferPtr++] << 8) | buffer[bufferPtr++]);
      }
      else
      {
         return (short) (buffer[bufferPtr++] | (buffer[bufferPtr++] << 8));
      }
   }

  /**
   * Un-marshals an unsigned long from the stream.
   *
   * @return  next four octets interpreted as an unsigned long
   */
   public int readUnsignedLong()
   {
      return readLong();
   }

  /**
   * Un-marshals an long from the stream.
   *
   * @return  next four octets interpreted as a signed long
   */
   public int readLong()
   {
      //Align bufferPtr so it points to the next multiple of 4.
      bufferPtr = align(bufferPtr, 4);
      if(bigEndian)
      {
         return (int) (((buffer[bufferPtr++]&0xFF) << 24) |
                       ((buffer[bufferPtr++]&0xFF) << 16) |
                       ((buffer[bufferPtr++]&0xFF) << 8)  |
                        (buffer[bufferPtr++]&0xFF)        );
      }
      else
      {
         return (int) ( (buffer[bufferPtr++]&0xFF)        |
                       ((buffer[bufferPtr++]&0xFF) << 8)  |
                       ((buffer[bufferPtr++]&0xFF) << 16) |
                       ((buffer[bufferPtr++]&0xFF) << 24) );
      }
   }

   /**
   * Un-marshals an unsigned long long from the stream.
   *
   * @return next eight bytes interpreted as an unsigned long long
   */
   public long readUnsignedLongLong()
   {
      return readLongLong();
   }

  /**
   * Un-marshals a long long from the stream
   *
   * @return next eight bytes interpreted as a signed long long
   */
   public long readLongLong()
   {
      long val = 0;
      long lval1;
      long lval2;

      //Align bufferPtr so it points to the next multiple of 8.
      bufferPtr = align(bufferPtr, 8);

      if(bigEndian)
      {
         lval1 = (long) (((buffer[bufferPtr++]&0xFF) << 24) |
                         ((buffer[bufferPtr++]&0xFF) << 16) |
                         ((buffer[bufferPtr++]&0xFF) <<  8) |
                          (buffer[bufferPtr++]&0xFF) );

         lval2 = (long) (((buffer[bufferPtr++]&0xFF) << 24) |
                         ((buffer[bufferPtr++]&0xFF) << 16) |
                         ((buffer[bufferPtr++]&0xFF) << 8)  |
                          (buffer[bufferPtr++]&0xFF) );

         val = (long) ( (lval1 << 32) | lval2 );
      }
      else
      {
         lval1 = (long)  ((buffer[bufferPtr++]&0xFF)        |
                         ((buffer[bufferPtr++]&0xFF) << 8)  |
                         ((buffer[bufferPtr++]&0xFF) << 16) |
                         ((buffer[bufferPtr++]&0xFF) << 24));

         lval2 = (long)  ((buffer[bufferPtr++]&0xFF)        |
                         ((buffer[bufferPtr++]&0xFF) << 8)  |
                         ((buffer[bufferPtr++]&0xFF) << 16) |
                         ((buffer[bufferPtr++]&0xFF) << 24));

         val = (long) ( (lval2 << 32) | lval1 );
      }
      return val;
   }

  /**
   * Un-marshals a float from the stream
   *
   * @return next four octets interpreted as a 32 bit floating point value
   */
   //public float readFloat()
   //{
      //return java.lang.Float.intBitsToFloat(readLong());
   //}

  /**
   * Un-marshals a double from the stream
   *
   * @return next eight octets interpreted as a 64 bit floating point value
   */
   //public double readDouble()
   //{
      //return java.lang.Double.longBitsToDouble(readLongLong());
   //}

   /**
    * Un-marshals a string from the stream
    *
    * @return String A string
    */
   public String readString()
   {
      int len = readLong();
      byte[] buf = new byte[len-1];
      System.arraycopy(buffer, bufferPtr, buf, 0, len-1);
      bufferPtr += len;
      return new String( buf );
   }

   public byte[] readOctetSequence()
   {
      int len = readLong();
      byte[] buf = new byte[len];
      System.arraycopy(buffer, bufferPtr, buf, 0, len);
      bufferPtr += len;
      return buf;
   }

   /**
    * Gets the buffer.
    *
    * @return  underlying buffer used to hold the CDR stream
    *
    */
   public byte[] getBuffer()
   {
      return buffer;
   }

   /**
    * Sets the buffer. This method automatically resets the CDR stream.
    *
    * @param buffer  underlying buffer used to hold the CDR stream
    *
    */
   public void setBuffer(byte[] buffer)
   {
      this.buffer = buffer;
   }

   /**
    * Resets this CDR stream causing it to start reading from the 
    * beginning of the stream.
    */
   public void reset()
   {
      bufferPtr = 0;
   }

   private int align( int ptr, int alignment )
   {
      if ( m_align )
      {
         return (ptr + alignment - 1) & ~(alignment-1);
      }
      else
      {
         return ptr;
      }
   }

}
