/*
 * @(#)MessageDigest.java   1.73 01/12/03
 *
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
/* Changes were made to this file in order to accomodate MQC's needs. */

package mqc.security;

import java.util.*;
import java.lang.*;
import java.io.IOException;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.InputStream;
import java.io.ByteArrayInputStream;

/**
 * This MessageDigest class provides applications the functionality of a
 * message digest algorithm, such as MD5 or SHA.
 * Message digests are secure one-way hash functions that take arbitrary-sized
 * data and output a fixed-length hash value.
 *
 * <p>A MessageDigest object starts out initialized. The data is 
 * processed through it using the {@link #update(byte) update}
 * methods. At any point {@link #reset() reset} can be called
 * to reset the digest. Once all the data to be updated has been
 * updated, one of the {@link #digest() digest} methods should 
 * be called to complete the hash computation.
 *
 * <p>The <code>digest</code> method can be called once for a given number 
 * of updates. After <code>digest</code> has been called, the MessageDigest
 * object is reset to its initialized state.
 *
 * <p>Implementations are free to implement the Cloneable interface.
 * Client applications can test cloneability by attempting cloning
 * and catching the CloneNotSupportedException: <p>    
 *
* <pre>
* MessageDigest md = MessageDigest.getInstance("SHA");
*
* try {
*     md.update(toChapter1);
*     MessageDigest tc1 = md.clone();
*     byte[] toChapter1Digest = tc1.digest();
*     md.update(toChapter2);
*     ...etc.
* } catch (CloneNotSupportedException cnse) {
*     throw new DigestException("couldn't make digest of partial content");
* }
* </pre>
 *
 * <p>Note that if a given implementation is not cloneable, it is
 * still possible to compute intermediate digests by instantiating
 * several instances, if the number of digests is known in advance.
 *
 * <p>Note that this class is abstract and extends from
 * <code>MessageDigestSpi</code> for historical reasons.
 * Application developers should only take notice of the methods defined in
 * this <code>MessageDigest</code> class; all the methods in
 * the superclass are intended for cryptographic service providers who wish to
 * supply their own implementations of message digest algorithms.
 *
 * @author Benjamin Renaud 
 *
 * @version 1.73, 12/03/01
 *
 * @see DigestInputStream
 * @see DigestOutputStream
 */

public abstract class MessageDigest
{   //extends MessageDigestSpi {

   //FIXME: include the algorithm name later (not supported by Cryptix impl.s)
   //private String algorithm;

   // The state of this digest
   private static final int INITIAL = 0;
   private static final int IN_PROGRESS = 1;
   private int state = INITIAL;

   //FIXME: include support for constructors with the algorithm name
   //    /**
   //     * Creates a message digest with the specified algorithm name.
   //     *
   //     * @param algorithm the standard name of the digest algorithm.
   //     * See Appendix A in the <a href=
   //     * "../../../guide/security/CryptoSpec.html#AppA">
   //     * Java Cryptography Architecture API Specification &amp; Reference </a>
   //     * for information about standard algorithm names.
   //     */
   //    protected MessageDigest(String algorithm) {
   //    this.algorithm = algorithm;
   //    }

   /**
    * Updates the digest using the specified byte.    
    * 
    * @param input the byte with which to update the digest.
    */
   public void update(byte input)
   {
      engineUpdate(input);
      state = IN_PROGRESS;
   }

   /**
    * Updates the digest using the specified array of bytes, starting
    * at the specified offset.
    * 
    * @param input the array of bytes.
    *
    * @param offset the offset to start from in the array of bytes.
    *
    * @param len the number of bytes to use, starting at 
    * <code>offset</code>.  
    */
   public void update(byte[] input, int offset, int len)
   {
      if (input == null)
      {
         throw new IllegalArgumentException("No input buffer given");
      }
      if (input.length - offset < len)
      {
         throw new IllegalArgumentException("Input buffer too short");
      }
      engineUpdate(input, offset, len);
      state = IN_PROGRESS;
   }

   /**
    * Updates the digest using the specified array of bytes.
    * 
    * @param input the array of bytes.
    */
   public void update(byte[] input)
   {
      engineUpdate(input, 0, input.length);
      state = IN_PROGRESS;
   }

   /**
    * Completes the hash computation by performing final operations
    * such as padding. The digest is reset after this call is made.
    *
    * @return the array of bytes for the resulting hash value.  
    */
   public byte[] digest()
   {
      /* Resetting is the responsibility of implementors. */
      byte[] result = engineDigest();
      state = INITIAL;
      return result;
   }

   /**
    * Completes the hash computation by performing final operations
    * such as padding. The digest is reset after this call is made.
    *
    * @param buf output buffer for the computed digest
    *
    * @param offset offset into the output buffer to begin storing the digest
    *
    * @param len number of bytes within buf allotted for the digest
    *
    * @return the number of bytes placed into <code>buf</code>
    * 
    * @exception DigestException if an error occurs.
    */
   public int digest(byte[] buf, int offset, int len) throws DigestException
   {
      if (buf == null)
      {
         throw new IllegalArgumentException("No output buffer given");
      }
      if (buf.length - offset < len)
      {
         throw new IllegalArgumentException ("Output buffer too small for specified offset and length");
      }
      int numBytes = engineDigest(buf, offset, len);
      state = INITIAL;
      return numBytes;
   }

   /**
    * Performs a final update on the digest using the specified array 
    * of bytes, then completes the digest computation. That is, this
    * method first calls {@link #update(byte[]) update(input)},
    * passing the <i>input</i> array to the <code>update</code> method,
    * then calls {@link #digest() digest()}.
    *
    * @param input the input to be updated before the digest is
    * completed.
    *
    * @return the array of bytes for the resulting hash value.  
    */
   public byte[] digest(byte[] input)
   {
      update(input);
      return digest();
   }

   //FIXME: this is rather useless without subclass support which doesn't exist
   //    /**
   //     * Returns a string representation of this message digest object.
   //     */
   //    public String toString() {
   //    ByteArrayOutputStream baos = new ByteArrayOutputStream();
   //    PrintStream p = new PrintStream(baos);
   //    p.print("Message Digest: ");
   //    switch (state) {
   //    case INITIAL:
   //        p.print("<initialized>");
   //        break;
   //    case IN_PROGRESS:
   //        p.print("<in progress>");
   //        break;
   //    }
   //    p.println();
   //    return (baos.toString());
   //    }

   /**
    * Compares two digests for equality. Does a simple byte compare.
    * 
    * @param digesta one of the digests to compare.
    * 
    * @param digestb the other digest to compare.    
    *
    * @return true if the digests are equal, false otherwise.
    */
   public static boolean isEqual(byte digesta[], byte digestb[])
   {
      if (digesta.length != digestb.length)
      {
         return false;
      }

      for (int i = 0; i < digesta.length; i++)
      {
         if (digesta[i] != digestb[i])
         {
            return false;
         }
      }
      return true;
   }

   /**
    * Resets the digest for further use.
    */
   public void reset()
   {
      engineReset();
      state = INITIAL;
   }

   //FIXME: add support for the algorithm name
   //    /**
   //     * Returns a string that identifies the algorithm, independent of
   //     * implementation details. The name should be a standard
   //     * Java Security name (such as "SHA", "MD5", and so on).
   //     * See Appendix A in the <a href=
   //     * "../../../guide/security/CryptoSpec.html#AppA">
   //     * Java Cryptography Architecture API Specification &amp; Reference </a>
   //     * for information about standard algorithm names.
   //     *
   //     * @return the name of the algorithm
   //     */
   //    public final String getAlgorithm() {
   //    return this.algorithm;
   //    }

   /**
    * Returns the length of the digest in bytes, or 0 if this operation is
    * not supported by the provider and the implementation is not cloneable.
    *
    * @return the digest length in bytes, or 0 if this operation is not
    * supported by the provider and the implementation is not cloneable.
    * 
    * @since 1.2
    */
   public final int getDigestLength()
   {
      int digestLen = engineGetDigestLength();
      if (digestLen == 0)
      {
         try
         {
            MessageDigest md = (MessageDigest)clone();
            byte[] digest = md.digest();
            return digest.length;
         }
         catch (CloneNotSupportedException e)
         {
            return digestLen;
         }
      }
      return digestLen;
   }

   /**
    * Get the Block Size (needed for constructing an HMAC from this MD
    */
   public final int getBlockSize()
   {
      return engineGetBlockSize();
   }

   // SPI methods
   //.............................

   public abstract Object clone() throws CloneNotSupportedException;

   protected abstract int engineGetDigestLength();

   protected abstract int engineGetBlockSize();

   protected abstract void engineUpdate(byte input);

   protected abstract void engineUpdate(byte[] input, int offset, int len);

   protected abstract byte[] engineDigest();

   protected abstract int engineDigest(byte[] buf, int offset, int len) throws DigestException;

   protected abstract void engineReset();
}
