/* $Id: Caesar.java,v 1.3 2003/05/28 03:35:15 enaess Exp $
 *
 * Copyright (C) 1995-2000 The Cryptix Foundation Limited.
 * All rights reserved.
 * 
 * Use, modification, copying and distribution of this software is subject 
 * the terms and conditions of the Cryptix General Licence. You should have 
 * received a copy of the Cryptix General Licence along with this library; 
 * if not, you can download a copy from http://www.cryptix.org/ .
 */
package mqc.security.cipher;

import mqc.security.Cipher;
//import java.security.AlgorithmParameters;
//import java.security.InvalidAlgorithmParameterException;

import mqc.security.InvalidKeyException;
import mqc.security.Key;
import mqc.security.NoSuchAlgorithmException;
import mqc.security.SecureRandom;
//import java.security.spec.AlgorithmParameterSpec;

import mqc.security.BadPaddingException;
//import javax.crypto.CipherSpi;
import mqc.security.IllegalBlockSizeException;
import mqc.security.NoSuchPaddingException;
import mqc.security.ShortBufferException;


/**
 * Caesar cipher (identity transformation).
 *
 * @version $Revision: 1.3 $
 * @author  Jeroen C. van Gelderen (gelderen@cryptix.org)
 */
public final class Caesar extends Cipher
{
   /**
    * The block size of this cipher. Being a stream cipher this value
    * is 1!
    */
   private static final int BLOCK_SIZE = 1;

   /** Raw Caesare key value. */
   private byte caesarKey;


   // Constructor
   //............................................................................

   public Caesar()
   {
      super("Caesar/ECB/NONE");
   }


   /**
    * Always throws a CloneNotSupportedException (cloning of ciphers is not
    * supported for security reasons).
    */
   public final Object clone() throws CloneNotSupportedException
   {
      throw new CloneNotSupportedException();
   }


   // Implementation of JCE methods
   //............................................................................

   protected final void engineSetMode(String mode)
      throws NoSuchAlgorithmException
   {
      if(mode.toUpperCase().equals("ECB"))
         return;
      else
         throw new NoSuchAlgorithmException();
   }


   protected final void engineSetPadding(String padding)
      throws NoSuchPaddingException
   {
      if(padding.toUpperCase().equals("NONE") ||
         padding.toUpperCase().equals("NOPADDING") )
      {
         return;
      }
      throw new NoSuchPaddingException();
   }


   /**
    * Returns the length of an input block, in bytes.
    *
    * @return the length in bytes of an input block for this cipher.
    */
   public int engineGetBlockSize ()
   {
      return BLOCK_SIZE;
   }


   /**
    * Return output size given input length. We don't buffer so this method
    * always returns inputLen.
    */
   protected final int engineGetOutputSize(int inputLen)
   {
      return inputLen;
   }


   /**
    * We don't use IVs.
    *
    * FIXME: maybe return whatever was set?
    */
   protected final byte[] engineGetIV()
   {
      return null;
   }


   protected final void engineInit(int opmode, Key key, SecureRandom random)
      throws InvalidKeyException
   {
      if( key==null )
         throw new InvalidKeyException("Key missing");

      if( !key.getFormat().toUpperCase().equals("RAW") )
         throw new InvalidKeyException("Wrong format: RAW bytes needed");

      byte[] userKey = key.getEncoded();
      if (userKey == null)
         throw new InvalidKeyException("RAW bytes missing");

      switch (opmode)
      {
      case Cipher.ENCRYPT_MODE:
         caesarKey = userKey[0];
         break;
      case Cipher.DECRYPT_MODE:
         caesarKey = (byte) -userKey[0];
         break;
      default:
         throw new InvalidKeyException("Unreachable code reached");
      }
   }


   protected final int
   engineUpdate(byte[] input, int inputOffset, int inputLen,
                byte[] output, int outputOffset)
      throws ShortBufferException
   {
      if( output.length-outputOffset < inputLen )
         throw new ShortBufferException();

      return this.internalUpdate(input, inputOffset, inputLen,
                                 output, outputOffset);
   }


   protected final byte[] engineUpdate(byte[] input, int inputOffset, int inputLen)
   {
      byte[] tmp  = new byte[this.engineGetOutputSize(inputLen)];
      this.internalUpdate(input, inputOffset, inputLen, tmp, 0);
      return tmp;
   }


   protected final int
   engineDoFinal(byte[] input, int inputOffset, int inputLen,
                 byte[] output, int outputOffset)
      throws ShortBufferException, IllegalBlockSizeException, BadPaddingException
   {
      return engineUpdate(input, inputOffset, inputLen, output, outputOffset);
   }


   protected final byte[]
   engineDoFinal(byte[] input, int inputOffset, int inputLen)
      throws IllegalBlockSizeException, BadPaddingException
   {
      byte[] tmp  = new byte[this.engineGetOutputSize(inputLen)];
      this.internalUpdate(input, inputOffset, inputLen, tmp, 0);
      return tmp;
   }


   private final int
   internalUpdate(byte[] input, int inputOffset, int inputLen,
                  byte[] output, int outputOffset)
   {
      System.arraycopy(input, inputOffset, output, outputOffset, inputLen);
      for (int i = outputOffset;  i < outputOffset + inputLen;  i++)
      {
         output[i] = (byte) (output[i] + caesarKey);
      }
      return inputLen;
   }
}
